package cz.frantovo.xmlWebGenerator;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.Date;
import java.net.URI;
import java.net.URISyntaxException;

/**
 * Knihovna funkcí volaných z XSLT.
 *  
 * TODO:
 *	- rozdělit na více modulů (jmenných prostorů).
 *	- CLI konektor
 * 
 * @author fiki
 */
public class Funkce {

	private static final String PŘÍKAZ_PYGMENTIZE = "pygmentize";
	private static final String PŘÍKAZ_DOT = "dot";
	private static final String PŘÍKAZ_WHICH = "which";
	private static final String ADRESÁŘ_VÝSTUPNÍ = "výstup";
	private static int počítadloDiagramů = 0;

	/**
	 * Zjištuje, kdy byl naposledy daný soubor změněn.
	 * @param soubor cesta k souboru
	 * @return datum poslední změny
	 * @throws URISyntaxException
	 */
	public static Date posledníZměna(String soubor) throws URISyntaxException {
		URI uri = new URI(soubor);
		File f = new File(uri);
		return new Date(f.lastModified());
	}

	/**
	 * Zvýrazňuje syntaxi zdrojového kódu. Používá k tomu externí program/knihovnu pygmentize.
	 * @param zdroják zdrojový kód, který předáme příkazu pygmentize na standardním vstupu
	 * @param jazyk předáme příkazu pygmentize jako parametr -l &lt;lexer&gt;
	 * @return zvýrazněný text nebo null, pokud došlo k chybě.
	 * TODO: 
	 *	- vracet místo textu instanci com.icl.saxon.om.NodeInfo http://saxon.sourceforge.net/saxon6.5.3/extensibility.html
	 *  - nebo kontrolovat validitu vygenerovaného kódu (v současnosti se spoléháme na bezchybnost pygmentize)
	 */
	public static String zvýrazniSyntaxi(String zdroják, String jazyk) throws IOException, InterruptedException {
		if (jazyk == null || jazyk.length() == 0) {
			System.err.println("Není vyplněn atribut „jazyk“ → není jasné, jak se má zvýrazňovat.");
			return null;
		} else if (isPříkazDostupný(PŘÍKAZ_PYGMENTIZE)) {
			Runtime r = Runtime.getRuntime();
			Process p = r.exec(new String[]{PŘÍKAZ_PYGMENTIZE, "-f", "html", "-l", jazyk});

			PrintStream vstupProcesu = new PrintStream(p.getOutputStream());
			vstupProcesu.print(zdroják);
			vstupProcesu.close();

			String výsledek = načtiProud(p.getInputStream());
			String chyby = načtiProud(p.getErrorStream());

			p.waitFor();

			if (chyby.length() == 0) {
				// Pozor: pygmentize má i při chybě návratový kód 0 → je potřeba kontrolovat chybový výstup.
				return výsledek;
			} else {
				System.err.print("Při zvýrazňování syntaxe došlo k chybě: " + chyby);
				return null;
			}
		} else {
			System.err.println("Příkaz " + PŘÍKAZ_PYGMENTIZE + " není na vašem systému dostupný → zvýrazňování syntaxe nebude fungovat.");
			System.err.println("Můžete ho nainstalovat pomocí:");
			System.err.println("\t$ aptitude install python-pygments   # (Debian/Ubuntu)");
			System.err.println("\t$ yum install python-pygments        # (Fedora/RedHat)");
			return null;
		}
	}

	/**
	 * Vygeneruje CSS styl pro zvýrazňování syntaxe.
	 * @return obsah CSS souboru nebo null, pokud generování nebylo možné
	 */
	public static String generujCssSyntaxe() throws IOException, InterruptedException {
		if (isPříkazDostupný(PŘÍKAZ_PYGMENTIZE)) {
			Runtime r = Runtime.getRuntime();
			Process p = r.exec(new String[]{PŘÍKAZ_PYGMENTIZE, "-S", "default", "-f", "html"});
			return načtiProud(p.getInputStream());
		} else {
			return null;
		}
	}

	/**
	 * Vytvoří obrázek s diagramem.
	 * @param zadání definice diagramu ve formátu dot
	 * @param vodorovně zda má být graf orientovaný vodorovně
	 * @return název souboru bez přípony, který byl vytvořen, nebo null, pokud došlo k chybě.
	 */
	public static String vytvořDiagram(String zadání, boolean vodorovně) throws IOException, InterruptedException {
		if (isPříkazDostupný(PŘÍKAZ_DOT)) {
			počítadloDiagramů++;
			String soubor = "diagram-" + počítadloDiagramů;
			String souborSložka = ADRESÁŘ_VÝSTUPNÍ + File.separator + soubor;

			StringBuilder zdroják = new StringBuilder(zadání.length() + 200);

			zdroják.append("digraph d {\n");
			zdroják.append("\tbgcolor=\"transparent\";\n");
			if (vodorovně) {
				zdroják.append("\trankdir=LR;");
			}
			zdroják.append(zadání);
			zdroják.append("}\n");

			Runtime r = Runtime.getRuntime();
			Process p = r.exec(new String[]{PŘÍKAZ_DOT, "-T", "svg", "-o", souborSložka + ".svg"});

			/**
			 * TODO: generovat i PNG bitmapu
			 */
			PrintStream vstupProcesu = new PrintStream(p.getOutputStream());
			vstupProcesu.print(zdroják.toString());
			vstupProcesu.close();

			String chyby = načtiProud(p.getErrorStream());

			p.waitFor();

			if (chyby.length() == 0) {
				return soubor;
			} else {
				System.err.print("Při vytváření diagramu došlo k chybě: " + chyby);
				return null;
			}
		} else {
			System.err.println("Příkaz " + PŘÍKAZ_DOT + " není na vašem systému dostupný → diagramy nelze vygreslit.");
			System.err.println("Můžete ho nainstalovat pomocí:");
			System.err.println("\t$ aptitude install graphviz   # (Debian/Ubuntu)");
			System.err.println("\t$ yum install graphviz        # (Fedora/RedHat)");
			return null;
		}
	}

	/**
	 * Čte proud dat dokud to jde a výsledek pak vrátí jako text.
	 * @param proud vstupní proud
	 * @return obsah proudu jako text
	 * @throws IOException 
	 */
	private static String načtiProud(InputStream proud) throws IOException {
		StringBuilder výsledek = new StringBuilder();
		BufferedReader buf = new BufferedReader(new InputStreamReader(proud));
		while (true) {
			String radek = buf.readLine();
			if (radek == null) {
				break;
			} else {
				výsledek.append(radek);
				výsledek.append("\n");
			}
		}
		return výsledek.toString();
	}

	/**
	 * Pomocí programu which zjistí, jestli je daný příkaz v systému přítomný.
	 * @param příkaz jehož přítomnost zjišťujeme
	 * @return true pokud příkaz v systému existuje
	 */
	private static boolean isPříkazDostupný(String příkaz) {
		try {
			Runtime r = Runtime.getRuntime();
			Process p = r.exec(new String[]{PŘÍKAZ_WHICH, příkaz});
			p.waitFor();
			return p.exitValue() == 0;
		} catch (Exception e) {
			System.err.printf("Při zjišťování dostupnosti příkazu „%s“ došlo k chybě: %s", příkaz, e.getLocalizedMessage());
			return false;
		}
	}
}
